import io
import csv
from datetime import datetime
from typing import Union
import sqlite3
import psycopg2
from sqlalchemy import Engine, create_engine, text

from .version import __version__
from .Series.series import Series
from .Dataset.dataset import Dataset
from .utils import to_datetime, date_range, parse_date

class Atrax:
    Series = Series
    DataSet = Dataset
    # qcut = qcut
    # cut = cut
    to_datetime = to_datetime
    date_range = date_range
    parse_date = parse_date
    version = __version__


    @staticmethod
    def read_csv(path_or_str, from_string=False, encoding='utf-8', converters=None, usecols=None, parse_dates=None):
        """Read a CSV file or string into a dataset.
        
        Parameters:
        -----------
            path_or_str: (str): Path to the CSV file or a CSV formatted string.
            from_string: (bool): If True, treats path_or_str as a string, otherwise as a file path.
            encoding: (str): Encoding to use when reading the file
            converters: (dict): Optional dict of colum: function
            usecols: (list): Optionaal list of columns to keep
            
        Returns:
        -----------
            DataSet: A DataSet object containing the data from the CSV.
        """
        if from_string:
            f = io.StringIO(path_or_str)
        else:
            f = open(path_or_str, newline='')

        reader = csv.DictReader(f)
        rows = []

        # attempt numeric conversion
        for row in reader:
            parsed_row = {}
            for k, v in row.items():
                if usecols and k not in usecols:
                    continue

                # Handle custom converter
                if converters and k in converters:
                    try:
                        parsed_row[k] = converters[k](v)
                        continue
                    except Exception:
                        parsed_row[k] = v
                        continue

                # Handle datetime parsing
                if parse_dates and k in parse_dates:
                    try:
                        parsed_row[k] = datetime.fromisoformat(v)
                        continue
                    except ValueError:
                        try:
                            parsed_row[k] = datetime.strptime(v, "%Y-%m-%d")
                        except:
                            parsed_row[k] = v
                        continue

                # Try numeric fallback
                try:
                    parsed_row[k] = float(v) if '.' in v else int(v)
                except:
                    parsed_row[k] = v

            rows.append(parsed_row)
                
        return Dataset(rows)
    
    @staticmethod
    def get_db(conn_str:str):
        """
        Get a database connection object from a connection string.

        Parameters:
            conn_str (str): Connection string for the database

        Returns:
            Database connection object
        """
        return create_engine(conn_str, echo=True)
    
    @staticmethod
    def read_sql(query: str, conn: Union[Engine, sqlite3.Connection, psycopg2.extensions.connection], index_col=None) -> "DataSet":
        """
        Execute SQL and convert result to Atrax DataSet, without using pandas.

        Parameters:
        -----------
        query : str
            SQL query to run.
        conn : SQLAlchemy engine/connection, sqlite3, or psycopg2 connection.
        index_col : str, optional
            Optional column to set as index.

        Returns:
        --------
        DataSet
        """
        try:
            # Handle SQLAlchemy connection
            if isinstance(conn, (Engine)):
                with conn.connect() as connection:
                    result = connection.execute(text(query))
                    columns = result.keys()
                    rows = result.fetchall()

            # Handle sqlite3 and psycopg2
            else:
                cursor = conn.cursor()
                cursor.execute(query)
                columns = [desc[0] for desc in cursor.description]
                rows = cursor.fetchall()

            # Convert rows into columns
            data = {col: [] for col in columns}
            for row in rows:
                for col, val in zip(columns, row):
                    data[col].append(val)

            # Set index if requested
            if index_col and index_col in data:
                index = data.pop(index_col)
                return Dataset(data, index=index_col, index_values=index)
            else:
                return Dataset(data)

        except Exception as e:
            raise RuntimeError(f"Failed to execute query: {e}")
    


from datetime import datetime
import numpy as np
from .indexers import _ILoc, _Loc
from ..utils.date_accessor import _DateTimeAccessor
import statistics

class Series:

    @property
    def iloc(self):
        """
            Provides integer-location based indexing

            Allows access to elements by thier integer position

            Examples:
            >>> from atrax import Series
            >>> s = Series([1,2,3], name='example', index=['a', 'b', 'c'])
            >>> s.iloc[0]
            1

            >>> s.iloc[1:3]
            b   2
            c   3
            Name: example, dtype: int64

            >>> s = Series([1,2,3,4,5,],  name="numbers", index=['a', 'b', 'c', 'd', 'e'])
            >>> s.iloc[1:4]
            b    2
            c    3
            d    4
            Name: numbers, dtype: int64   

            >>> s.iloc[::-1]
            e    5
            d    4
            c    3
            b    2
            a    1
            Name: numbers, dtype: int64                    
        """
        return _ILoc(self)
    
    @property
    def loc(self):
        """ 
            Provides label-based indexing for the Series
        
            Examples:

            >>> from atrax import Series
            >>> s = Series([1,2,3], name="example" index=['a', 'b', 'c'])
            >>> s.loc['a']
            1

            >>> s.loc['b': 'c']
            b   2
            c   3
            Name: example, dtype: int64
            
        """
        return _Loc(self)
    
    @property
    def values(self):
        """Returns the underlying data of the Series as a list.
        
        Examples:
        >>> from atrax import Atrax as tx
        >>> s = tx.Series([1, 2, 3], name='example', index=['a', 'b', 'c'])
        >>> s.values
        [1, 2, 3]
        
        >>> type(s.values)
        list"""
        return self.data

    @property
    def name(self):
        """Returns the name of the Series.

        Examples:
        >>> from atrax import Atrax as tx
        >>> s = tx.Series([1, 2, 3], name='example', index=['a', 'b', 'c'])
        >>> s.name
        'example'

        >>> s = tx.Series([1, 2, 3])
        >>> s.name
        ''
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value      
    
    @property
    def dt(self):
        """
        Provides datetime-like properties for the Series.
        
        Examples:
        >>> from atrax import Atrax as tx
        >>> test_data = [
            {
                'id': 1,
                'sale_date': '1/1/2025'
            },
            {
                'id': 2,
                'sale_date': '1/2/2025'
            },
            {
                'id': 3,
                'sale_date': '1/3/2025'
            }
        ]
        >>> ds = tx.DataSet(test_data)

        >>> ds['weekday'] = ds['sale_date'].dt.weekday
        >>> ds.head()
        id    sale_date    weekday
        1     1/1/2025     2
        2     1/2/2025     3
        3     1/3/2025     4


        >>> ds['is_weekend'] = ds['sale_date'].dt.is_weekend
        >>> ds.head()
        id    sale_date    weekday   is_weekend
        1     1/1/2025     2         False
        2     1/2/2025     3         False
        3     1/3/2025     4         False

        >>> ds['month'] = ds['sale_date'].dt.month
        >>> ds.head()
        id    sale_date    weekday   is_weekend   month
        1     1/1/2025     2         False        1
        2     1/2/2025     3         False        1
        3     1/3/2025     4         False        1

        >>> ds['day'] = ds['sale_date'].dt.day
        >>> ds.head()
        id    sale_date    weekday   is_weekend   month  day
        1     1/1/2025     2         False        1      1
        2     1/2/2025     3         False        1      2
        3     1/3/2025     4         False        1      3

        >>> ds['year'] = ds['sale_date'].dt.year
        >>> ds.head()
        id    sale_date    weekday   is_weekend   month  day  year
        1     1/1/2025     2         False        1      1    2025
        2     1/2/2025     3         False        1      2    2025
        3     1/3/2025     4         False        1      3    2025  


        """
        return _DateTimeAccessor(self)
    
    @property
    def dtype(self):
        return self._dtype
    
    

    def __init__(self, data, name=None, index=None):
        """
            One-dimensional labeled array for atrax

            Parameters
            _________
                data (list): a list of values
                name (str): name of the Series
                index (list): list of indexes

            Examples:

            >>> from atrax import Series
            >>> s = Series([1, 2, 3])
            >>> s
            0    1
            1    2
            3    3
            Name: , dtype: int

            >>> s = tx.Series([1, 2, 3, 4, 5], name='numbers', index=['a', 'b', 'c'])
            >>> s
            a     1
            b     2  
            c     3
            Name: numbers, dtype: int

            >>> s = tx.Series([1.0, 2.0, 3.0], name='example', index=['a', 'b', 'c'])
            a     1.0
            b     2.0
            c     3.0
            Name: example, dtype: float

            >>> s = tx.Series(['hello', 'goodbye', 'whatsup'])
            0     hello
            1     goodbye
            2     whatsup
            Name: , dtype: str    

            >>> s = tx.Series([1, True, 'sexy', 2.5])
            0     1
            1     True
            2     sexy
            3     2.5
            Name: , dtype: object

            ##### this one is interesting and probably needs attention
            >>> s = tx.Series([True, False, True])
            0    True
            1    False
            2    True
            Name: , dtype: int                     
        """
        self.data = data
        self.name = name or ""
        # if the user didn't specify indexes, then we can infer them from the length of the data
        self.index = index or list(range(len(data)))
        # let's make sure that if the user supplied indexes that the length is correct
        if len(self.data) != len(self.index):
            raise ValueError(f"Length of index must match the length of the data, you supplied {len(self.data)} for data and {len(self.index)} for indexes")
        self._dtype = self._infer_dtype()

    def __len__(self):
        return len(self.data)
    
    def __iter__(self):
        return iter(self.data)

    def __repr__(self):
        lines = []
        lines.append(f'<Series />')
        lines.extend([f'{str(i)}    {str(d)}' for i, d in zip(self.index[0:10], self.data[0:10])])
        if len(self.data) > 10:
            lines.append(f"...({len(self.data)} total items)")
        lines.append(f'Name: {self.name}   dtype: {self._dtype}')
        return '\n'.join(lines)

    def _repr_html_(self):
        html = "<table style='border-collapse: collapse;'>"
        for idx, val in zip(self.index[:10], self.data[:10]):
            html += f"<tr><td style='text-align: 'left';'>{idx}</td>"
            html += f"<td>{val}</td></tr>"
        html += f"<tr><td colspan='2' style='font-size:14px;'>Name: {self.name}, dtype: {self._dtype}</td></tr>"
        if len(self.data) > 10:
            html += f"<tr><td colspan='2'><i>...{len(self.data)-10} more</i></td></tr>"
        html += "</table>"
        return html
        

    def _infer_dtype(self):
        if all(isinstance(x, float) for x in self.data):
            return 'float64'
        elif all(isinstance(x, bool) for x in self.data):
            return 'bool'
        elif all(isinstance(x, int) for x in self.data):
            return 'int64'
        elif all(isinstance(x, str) for x in self.data):
            return 'str'
        elif all(isinstance(x, datetime) for x in self.data):
            return 'datetime'
        else:
            return 'object'
        
    def _binary_op(self, other, op):
        if isinstance(other, Series):
            if len(self.data) != len(other.data):
                raise ValueError(f"Series must have the same length. found: {len(self.data)} and {len(other.data)}")
            return Series([op(a,b) for a,b in zip(self.data, other.data)], name=self.name)
        else:
            return Series([op(a, other) for a in self.data], name=self.name)
        
    # basic math operators
    def __add__(self, other): 
        return self._binary_op(other, lambda a, b: a + b)
    
    def __sub__(self, other): 
        return self._binary_op(other, lambda a,b: a - b)
    
    def __mul__(self, other): 
        return self._binary_op(other, lambda a,b: a * b)
    
    def __truediv__(self, other): 
        return self._binary_op(other, lambda a, b: a / b)
    
    def __floordiv__(self, other): 
        return self._binary_op(other, lambda a, b: a // b)
    
    def __mod__(self, other): 
        return self._binary_op(other, lambda a, b: a % b)
    
    def __pow__(self, other): 
        return self._binary_op(other, lambda a, b: a ** b)
    
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        from numpy import ndarray
        raw_inputs = []
        for x in inputs:
            if isinstance(x, Series):
                raw_inputs.append(x.data)
            else:
                raw_inputs.append(x)

        result = getattr(ufunc, method)(*raw_inputs, **kwargs)

        if isinstance(result, tuple):
            # for ufuncs that return multiple outputs, return tuple of series
            return tuple(Series(r, name=self.name) if isinstance(r, ndarray) else r for r in result)
        elif isinstance(result, ndarray):
            return Series(result.tolist(), name=self.name)
        else:
            return result

        
    def head(self, n=5):
        """
        Return the first n elements of the Series.

        Parameters:
        n (int): The number of elements to return. Defaults to 5.


        Returns:
        Series: A new Series containing the first n elements.

        Example usage:
        >>> s = tx.Series([1, 2, 3, 4, 5])
        >>> print(s.head(3))
        0    1
        1    2
        2    3
        Name: , dtype: int
        """        
        return Series(self.data[:n], name=self.name, index=self.index[:n])
    
    def tail(self, n=5):
        """
        Return the last n elements of the Series.

        Parameters:
        n (int): The number of elements to return. Defaults to 5.

        Returns:
        Series: A new Series containing the last n elements.

        Example usage:
        >>> s = tx.Series([1, 2, 3, 4, 5])
        >>> print(s.tail(3))
        2    3
        3    4 
        4    5 
        """        
        return Series(self.data[-n:], name=self.name, index=self.index[-n:])
    
    # reverse math operators
    def __radd__(self, other):
        return self._binary_op(other, lambda a, b: b + a)
    
    def __rsub__(self, other): 
        return self._binary_op(other, lambda a,b: b - a)
    
    def __rmul__(self, other): 
        return self._binary_op(other, lambda a,b: b * a)
    
    def __rtruediv__(self, other): 
        return self._binary_op(other, lambda a, b: b / a)
    
    def __rfloordiv__(self, other): 
        return self._binary_op(other, lambda a, b: b // a)
    
    def __rmod__(self, other): 
        return self._binary_op(other, lambda a, b: b % a)
    
    def __rpow__(self, other): 
        return self._binary_op(other, lambda a, b: b ** a)    
    
    # comparisons
    def __gt__(self, other):
        return Series([x > other for x in self.data], name=self.name)
    
    def __lt__(self, other):
        return Series([x < other for x in self.data], name=self.name)
    
    def __ge__(self, other):
        return Series([x >= other for x in self.data], name=self.name)
    
    def __le__(self, other):
        return Series([x <= other for x in self.data], name=self.name)
    
    def __eq__(self, other):
        return Series([x == other for x in self.data], name=self.name)
    
    def __ne__(self, other):
        return Series([x != other for x in self.data], name=self.name)
        
    # chaining
    def __and__(self, other):
        if not isinstance(other, Series):
            raise TypeError('Operand must be a Series')
        if len(other) != len(self.data):
            raise ValueError('Cannot perform operation. Series must have the same length')
        return Series([a and b for a,b in zip(self.data, other.data)], name=self.name)  

    def __or__(self, other):
        if not isinstance(other, Series):
            raise TypeError('Operand must be a Series')
        if len(other) != len(self.data):
            raise ValueError('Cannot perform operation. Series must have the same length')        
        return Series([a or b for a, b in zip(self.data, other.data)], name=self.name)      
    
    # unique
    def unique(self):
        """
        Return the unique values in the Series.

        Returns:
        Series: A new Series containing the unique values.

        Example usage:
        >>> s = tx.Series([1, 2, 2, 3, 4, 4])
        >>> unique_s = s.unique()
        >>> print(unique_s)
        0    1
        1    2
        2    3
        3    4
        Name: Unique(), dtype: int

        """        
        # the best way to do this is to create a set to remove duplicates
        uniques = list(set(self.data))
        # grab the unique values and calculate new indexes
        #return Series(uniques, name=self.name, index=list(range(len(uniques))))
        return np.array(uniques)
    
    # nunique
    def nunique(self):
        """
        Return the number of unique values in the Series.

        Returns:
        int: The number of unique values.

        Example usage:
        >>> s = tx.Series([1, 2, 2, 3, 4, 4])
        >>> num_unique = s.nunique()
        >>> print(num_unique)
        4

        """        
        uniques = list(set(self.data))
        return len(uniques)

    # apply
    def apply(self, func):
        """Apply a function to each element in the Series.
        
        Parameters:
        func (function): A function to apply to each element.   
        
        Returns:
        Series: A new Series with the function applied to each element.
        
        Example usage:
        >>> s = tx.Series([1, 2, 3])
        >>> result = s.apply(lambda x: x * 2)
        >>> print(result)
        0    2
        1    4
        2    6
        Name: , dtype: int
        
        >>> def square(x):
        >>>     return x** 2
            
        >>> result = s.apply(square)   
        >>> result
        0    1
        1    4  
        2    9
        Name: , dtype: int
        """        
        result = [func(d) for d in self.data]
        return Series(result, name=self.name)
    
    # map
    def map(self, arg):
        """
            Map values of the Series using an input mapping or function

            Parameters:

                arg (dict or function): a mapping dictionary or function to apply to each value
            
            Returns:

                Series

            Example usage:

            >>> s = Series([1,2,3], name='example', index=['a', 'b', 'c'])
            >>> mapping = {1: 'one', 2: 'two', 3: 'three'}
            >>> mapped_s = s.map(mapping)
            >>> mapped_s
            a     one
            b     two
            c     three
            Name: example, dtype: str
        """
        if callable(arg):
            mapped = [arg(x) for x in self.data]
        elif isinstance(arg, dict):
            mapped = [arg.get(x, None) for x in self.data]
        else:
            raise TypeError("Argument must be a callable of dictionary")
        return Series(mapped, name=self.name,index=self.index)
    
    # astype
    def astype(self, dtype):
        """
        Convert the Series to a specified data type.
        
        Parameters:

        dtype (type): The Python type to cast to (e.g., int, float, str)
        
        Returns:

        Series: A new Series with the converted data type.

        Example usage:
        >>> s = tx.Series(['1', '2', '3'])
        >>> i_series = s.astype(int)
        >>> i_series
        0    1
        1    2
        2    3
        Name: , dtype: int

        >>> s = tx.Series(['1', '2', '3'])
        >>> i_series = s.astype('int')
        >>> i_series
        0    1
        1    2
        2    3
        Name: , dtype: int       

        >>> s = tx.Series(['1', '2', '3'])
        >>> f_series = s.astype('float')
        >>> f_series
        0    1.0
        1    2.0
        2    3.0
        Name: , dtype: float           
        """        
        type_map = {
            'int': int,
            'float': float,
            'str': str,
            'object': lambda x: x
        }

        if isinstance(dtype, str):
            cast_fn = type_map.get(dtype)
            if cast_fn is None:
                raise ValueError(f"Unsupported dtype: {dtype}")
        else:
            cast_fn = dtype
        
        new_data = []
        for val in self.data:
            try:
                new_data.append(cast_fn(val))
            except:
                new_data.append(None)

        return Series(new_data, name=self.name, index=self.index)
    
    def to_list(self):
        """
        Convert the Series to a list.

        Returns:
        list: The data in the Series as a list.

        Example usage:
        >>> s = tx.Series([1, 2, 3])
        >>> lst = s.to_list()
        >>> print(lst)
        [1, 2, 3]

        """
        return self.data    
    
    def to_array(self):
        return self.data
    
    def __array__(self, dtype=None):
        return np.asarray(self.data, dtype=dtype)
    
    def mean(self, axis=None):
        return sum(self.data) / len(self.data)
    
    def std(self, axis=None, ddof=0):
        return statistics.stdev(self.data)
    
    def var(self, sample=True):
        n = len(self.data)
        if n < 2 and sample:
            raise ValueError("Sample variance requires at least two data points")
        mean = sum(self.data) / n
        sum_sq_diff = sum((x - mean) ** 2 for x in self.data)
        return sum_sq_diff / ( n - 1 if sample else n)

    def sum(self, axis=None):
        return sum(self.data)

    def min(self, axis=None):
        return min(self.data)

    def max(self, axis=None):
        return max(self.data)

    def median(self, axis=None):
        return statistics.median(self.data)

    def prod(self, axis=None):
        result = 1
        for x in self.data:
            result *= x
        return result

    def cumsum(self, axis=None):
        result = []
        total = 0
        for x in self.data:
            total += x
            result.append(total)
        return Series(result, name=self.name)

    def cumprod(self, axis=None):
        result = []
        total = 1
        for x in self.data:
            total *= x
            result.append(total)
        return Series(result, name=self.name)

    def cummin(self):
        result = []
        current_min = None
        for x in self.data:
            current_min = x if current_min is None else min(current_min, x)
            result.append(current_min)
        return Series(result, name=self.name)

    def cummax(self):
        result = []
        current_max = None
        for x in self.data:
            current_max = x if current_max is None else max(current_max, x)
            result.append(current_max)
        return Series(result, name=self.name)   



from datetime import datetime, timedelta
from typing import Union, List, Optional, Literal

def parse_date(val, fmt: Optional[str] = None) -> datetime:
    """Parse a date string or datetime object into a datetime object."""
    if isinstance(val, datetime):
        return val
    if isinstance(val, str):
        if fmt:
            return datetime.strptime(val, fmt)
        
        fallback_formats = ["%Y-%m-%d", "%m/%d/%Y", "%m/%d/%y", "%m-%d-%Y"]
        for try_fmt in fallback_formats:
            try:
                return datetime.strptime(val, try_fmt)
            except ValueError:
                continue

        try:
            return datetime.fromisoformat(val)
        except ValueError:
            pass
    raise TypeError(f"Could not parse date: {val!r}. Provide a valid date string or datetime object")



def to_datetime(
    values: Union[str, datetime, List[Union[str, datetime]]],
    fmt: Optional[str] = None,
    errors: str = "raise"  # or 'coerce'
) -> Union[datetime, List[Optional[datetime]]]:
    """
    Convert a string, datetime, or list of them to datetime objects.

    Parameters:
        values (str | datetime | list): Input date(s) to convert.
        fmt (str, optional): Optional format string to use for parsing.
        errors (str): If 'raise', invalid parsing raises an error.
                      If 'coerce', invalid parsing returns None.

    Returns:
        datetime | list[datetime | None]: Parsed datetime(s).

    Examples:
        >>> to_datetime("2023-10-01")
        datetime.datetime(2023, 10, 1)

        >>> to_datetime(["2023-10-01", "2023/10/02"])
        [datetime.datetime(2023, 10, 1), datetime.datetime(2023, 10, 2)]

        >>> to_datetime("01-10-2023", fmt="%d-%m-%Y")
        datetime.datetime(2023, 10, 1)
    """
    
    fallback_formats = ("%Y-%m-%d", "%m/%d/%Y", "%m/%d/%y", "%m-%d-%Y")

    def parse_single(val) -> Optional[datetime]:
        if val in (None, "", "NaT"):
            return None if errors == "coerce" else _raise(val)
        try:
            return parse_date(val, fmt=fmt)
        except Exception:
            return None if errors == "coerce" else _raise(val)

    def _raise(val):
        raise ValueError(f"Cannot parse date: {val}")

    # Series support
    from atrax.Series.series import Series
    if isinstance(values, Series):
        try:
            values = list(values)  # if iterable
        except TypeError:
            values = values.data  # fallback if needed

    if values is None:
        if errors == 'coerce':
            return None
        else:
            raise TypeError("Cannot parse None value without errors='coerce'")
        
    # Single value
    if isinstance(values, (str, datetime)):
        return parse_single(values)

    # List of values
    elif isinstance(values, list):
        return [parse_single(v) for v in values]

    else:
        raise TypeError(f"Unsupported input type: {type(values)}")
    
def date_range(
    start: Union[str, datetime],
    end: Optional[Union[str, datetime]] = None,
    periods: Optional[int] = None,
    freq: Literal['D', 'H', 'T', 'min', 'S'] = 'D',
    fmt: Optional[str] = None
) -> list[datetime]:
    """Generate a list of datetime values."""
    start_dt = parse_date(start, fmt)

    delta_map = {
        'D': timedelta(days=1),
        'H': timedelta(hours=1),
        'T': timedelta(minutes=1),
        'min': timedelta(minutes=1),
        'S': timedelta(seconds=1)
    }

    if freq not in delta_map:
        raise ValueError(f"Unsupported frequency: {freq}. Supported: {list(delta_map.keys())}")

    delta = delta_map[freq]

    if end is not None:
        end_dt = parse_date(end, fmt)
        if start_dt > end_dt:
            raise ValueError("'start' must be before 'end'.")
        result = []
        current = start_dt
        while current <= end_dt:
            result.append(current)
            current += delta
        return result

    if periods is not None:
        if periods <= 0:
            raise ValueError("'periods' must be a positive integer.")
        return [start_dt + i * delta for i in range(periods)]

    raise ValueError("Either 'end' or 'periods' must be specified.")





class _ILoc:
    def __init__(self, series):
        self.series = series

    def __getitem__(self, key):
        from .series import Series
        if isinstance(key, slice):
            return Series(self.series.data[key], name=self.series.name, index=self.series.index[key])
        else:
            return self.series.data[key]
        
class _Loc:
    def __init__(self, series):
        self.series = series

    def __getitem__(self, key):
        from .series import Series
        if isinstance(key, list):
            index_map = {k:v for k, v in zip(self.series.index, self.series.data)}
            return Series([index_map[k] for k in key], name=self.series.name, index=key)
        elif isinstance(key, slice):
            start_label = key.start
            end_label = key.stop

            try:
                start_idx = self.series.index.index(start_label)
            except ValueError:
                raise KeyError(f'start label not found in index: {start_label}')
            
            try:
                end_idx = self.series.index.index(end_label)
            except ValueError:
                raise KeyError(f'end label not found in index: {end_label}')
            
            data = self.series.data[start_idx:end_idx+1]
            index = self.series.index[start_idx:end_idx+1]
            return Series(data, name=self.series.name, index=index)
        else:
            idx = self.series.index.index(key)
            return self.series.data[idx]


from datetime import datetime
import statistics
from .indexers import _iLocIndexer, _LocIndexer
from ..Series.series import Series
from copy import deepcopy
from collections.abc import Sequence
from .group import GroupBy

class Dataset:

    @property
    def loc(self):
        return _LocIndexer(self)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    def __init__(self, data: list[dict], index=None):
        """
            Initialize a dataset.

            Parameters:

                data (list[dict] or dict[list]): either row oriented or column oriented data 
        """
        if isinstance(data, dict):
            if len(set([len(v) for v in data.values()])) != 1:
                raise ValueError("All columns must have the same length")
            
            keys = list(data.keys())
            values = zip(*data.values())
            data = [dict(zip(keys, row)) for row in values]

        self.data = data
        self.columns = list(data[0].keys()) if data else []
        self._index_name = ''
        self._index = index if index is not None else list(range(len(data)))

    def __repr__(self):
        lines = []
        lines.append('<Dataset />')
        lines.extend([', '.join(self.columns)])
        for row in self.data[:10]:
            lines.append(', '.join(str(row.get(col, '')) for col in self.columns))
        if len(self.data) > 10:
            lines.append(f"...({len(self.data)}) total rows")
        return '\n'.join(lines)
    
    def _repr_html_(self):
        if not self.data:
            return '<i>Empty Dataset</i>'
        
        from io import StringIO
        import csv, base64
        
        headers = self.columns.copy()
        show_index = self._index is not None

        # new
        if show_index and self._index_name in headers:
            headers.remove(self._index_name)
        # Header row
        index_label = self._index_name if self._index_name else ""
        header_html = f"<th>{index_label}</th>" if show_index else ""
        header_html += ''.join(f"<th>{col}</th>" for col in headers)   

        # Data rows
        body_html = ''
        for i, row in enumerate(self.data):
            idx_val = self._index[i] if self._index else i
            if isinstance(idx_val, datetime):
                idx_str = idx_val.strftime('%y-%m-%d')
            else:
                idx_str = str(idx_val)
            row_html = f"<td><strong>{idx_str}</strong></td>" if show_index else ""
            row_html += ''.join(f"<td>{row.get(col, '')}</td>" for col in headers)
            body_html += f"<tr>{row_html}</tr>" 

        # CSV export as base64
        def to_csv_string():
            buffer = StringIO()
            writer = csv.writer(buffer)
            if show_index:
                writer.writerow([index_label] + headers)
            else:
                writer.writerow(headers)
            for i, row in enumerate(self.data):
                row_vals = [row.get(col, "") for col in headers]
                if show_index:
                    row_vals = [self._index[i] if self._index else i] + row_vals
                writer.writerow(row_vals)
            return buffer.getvalue()  

        csv_bytes = to_csv_string().encode("utf-8")
        b64 = base64.b64encode(csv_bytes).decode("utf-8")
        download_link = f'''<a download="dataset.csv" href="data:text/csv;base64,{b64}" target="_blank"
            style="
               display: inline-block;
               color: white;
               text-decoration: none;
               font-family: sans-serif;
           "
        >Download CSV</a>'''

        # Final HTML output
        return f"""
            <div>
                {download_link}
                <table>
                    <thead><tr>{header_html}</tr></thead>
                    <tbody>{body_html}</tbody>
                </table>
            </div>
        """   

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data[key]
        elif isinstance(key, str):
            return Series([row.get(key) for row in self.data], name=key)
        elif isinstance(key, Series) and all(isinstance(val, bool) for val in key.data):
            # filter rows using a boolean series
            if len(key.data) != len(self.data):
                raise ValueError("Boolean Series must match the length of the dataset")
            
            filtered = [row for row, flag in zip(self.data, key.data) if flag]
            return Dataset(filtered)
        
        elif isinstance(key, list):
            # column subset
            return Dataset([{k: row[k] for k in key if k in row} for row in self.data])
        
        else:
            raise TypeError("Key must be a string (column), list of strings (subset) or Series (boolean mask)")
        
    def __setitem__(self, key, value):
        if isinstance(value, Series):
            if len(value.data) != len(self.data):
                raise ValueError("Series length must match Dataset length")
            for row, val in zip(self.data, value.data):
                row[key] = val
        elif isinstance(value, list):
            if len(value) != len(self.data):
                raise ValueError("List length must match Dataset length")
            for row, val in zip(self.data, value):
                row[key] = val
        elif callable(value):
            for row in self.data:
                row[key] = value(row)
        elif isinstance(value, (int, float, str, bool)):
            # broadcast scalar value
            for row in self.data:
                row[key] = value
        else:
            raise TypeError(f"Cannot assign value of type {type(value)} to column '{key}'")

        if key not in self.columns:
            self.columns.append(key)
    
    def head(self, n=5):
        """ Return the first n rows of the dataset."""
        return Dataset(self.data[:n])
    
    def tail(self, n = 5):
        """ Return the last n rows of the dataset."""
        return Dataset(self.data[-n:], index=self._index[-n:])
    
    def shape(self):
        return (len(self.data), len(self.columns))
    
    
    def describe(self, numeric_only=False):
        numeric_cols = {
            col: [row[col] for row in self.data if isinstance(row.get(col), (int, float))] for col in self.columns
        }
        if numeric_only:
            numeric_cols = {
                col: [row[col] for row in self.data if isinstance(row.get(col), (int, float))]
                for col in self.columns
                if any(isinstance(row.get(col), (int, float)) for row in self.data)
            }

        summary_rows = []

        def percentile(data, p):
            data = sorted(data)
            idx = int(round(p * (len(data) - 1)))
            return data[idx]
        
        for stat in ['mean', 'std', 'min', 'Q1', 'median', 'Q3', 'max', 'count']:
            row={'stat': stat}
            for col, values in numeric_cols.items():
                if not values:
                    row[col] = None
                    continue
                if stat == 'mean':
                    row[col] = round(statistics.mean(values), 2)
                elif stat == 'std':
                    row[col] = round(statistics.stdev(values), 2) if len(values) > 1 else 0.0
                elif stat == 'min':
                    row[col] = min(values)
                elif stat == 'Q1':
                    row[col] = percentile(values, 0.25)
                elif stat == 'median':
                    row[col] = round(statistics.median(values), 2)
                elif stat == 'Q3':
                    row[col] = percentile(values, 0.75)
                elif stat == 'max':
                    row[col] = max(values)
                elif stat == 'count':
                    row[col] = len(values)
            summary_rows.append(row)
        
        return Dataset(summary_rows)
    
    def info(self):
        print(f"<class 'Dataset'>")
        print(f"Range Index: {len(self.data)} entries")
        print(f"Data columns (total {len(self.columns)} columns):")
        if not self.data:
            print("   No data available")
            return

        # if self._index_name and self._index:
        #     index_sample = self._index[0]
        #     if isinstance(index_sample, datetime):
        #         dtype = 'datetime'
        #     elif isinstance(index_sample, bool):
        #         dtype = 'bool'
        #     elif isinstance(index_sample, int):
        #         dtype = 'int'
        #     elif isinstance(index_sample, float):
        #         dtype = 'float'
        #     elif isinstance(index_sample, str):
        #         dtype = 'str'
        #     else:
        #         dtype = type(index_sample).__name__

        #     print(f"Index")
        #     print(f"    name: {self._index_name}")
        #     print(f"    dtype: {dtype}")
        #     print("")

        # now print the column info
        col_stats = {}

        for col in self.columns:
            values = [row.get(col) for row in self.data]
            non_nulls = [v for v in values if v is not None]

            sample = non_nulls[0] if non_nulls else None
            dtype = 'unknown'

            if sample is None:
                dtype = 'NoneType'
            elif isinstance(sample, bool):
                dtype = 'bool'
            elif isinstance(sample, int):
                dtype = 'int'
            elif isinstance(sample, float):
                dtype = 'float'
            elif isinstance(sample, datetime):
                dtype = 'datetime'
            elif isinstance(sample, str):
                dtype = 'str'

            col_stats[col] = {
                'dtype': dtype,
                'non_null': len(non_nulls),
                'total': len(values)
            }
        print(f"{'Column':<15} | {'Dtype':<10} | {'Non-Null':<10} | {'Total':<10}")
        print(f"-" * 50)
        for col, stats in col_stats.items():
            print(f"{col:<15} | {stats['dtype']:<10} | {stats['non_null']:<10} | {stats['total']:<10}")

    def set_index(self, column, *, inplace=True, drop=True):
        """
        Set (or compute) the row index for this DataSet.

        Parameters
        ----------
        column : str | callable | Sequence
            • str   - name of an existing column to promote to the index  
            • callable - a function `f(row_dict) -> hashable` evaluated per row  
            • list/seq - explicit index values (must match len(self))
        inplace : bool, default True
            If True, mutate this DataSet.  If False, return a **new** DataSet.
        drop : bool, default True
            When `column` is a str: drop that column from the data rows.
            I think we mostly want this, so our display looks cleaner.

        Returns
        -------
        DataSet | None
            The updated DataSet (if `inplace=False`), otherwise ``None``.
        """
        from copy import deepcopy
        from collections.abc import Sequence
        from atrax import Dataset
        from atrax import Series

        # ------------------------------------------------------------------ #
        # 1. Determine the new index values (index_vals) and whether we have
        #    a column name that *might* need to be dropped.
        # ------------------------------------------------------------------ #
        if isinstance(column, str):
            if column not in self.columns:
                raise KeyError(f"Column '{column}' not found in dataset.")

            index_vals = [row[column] for row in self.data]
            index_name = column
            need_drop  = drop

        elif callable(column):
            index_vals = [column(row) for row in self.data]
            index_name = None        # anonymous index
            need_drop  = False       # nothing to drop

        elif isinstance(column, (list, tuple, Series)) or (
                isinstance(column, Sequence) and not isinstance(column, str)):
            # Accept list / tuple / Series / numpy array … anything Sequence-ish
            if len(column) != len(self.data):
                raise ValueError("Index length must match number of rows.")
            index_vals = list(column)          # materialise
            index_name = None
            need_drop  = False

        else:
            raise TypeError(
                "`column` must be a string, callable, or sequence of index values."
            )

        # ------------------------------------------------------------------ #
        # 2. Prepare the new data/columns containers
        # ------------------------------------------------------------------ #
        if inplace:
            target = self
            if need_drop:
                # remove the column from each row
                for row in target.data:
                    row.pop(column, None)
        else:
            target        = Dataset(deepcopy(self.data))  # deep-copy rows
            if need_drop:
                for row in target.data:
                    row.pop(column, None)

        # Refresh columns list if we dropped something
        if need_drop:
            target.columns = list(target.data[0].keys()) if target.data else []

        # ------------------------------------------------------------------ #
        # 3. Assign index + index name
        # ------------------------------------------------------------------ #
        target._index       = index_vals
        target._index_name  = index_name

        # ------------------------------------------------------------------ #
        # 4. Return as per `inplace`
        # ------------------------------------------------------------------ #
        if inplace:
            return None
        else:
            return target   

    def drop(self, columns=None, index=None, inplace=False):
        """
        Remove columns and/or rows.

        Parameters
        ----------
        columns : str | Sequence[str] | None
            Column name(s) to drop.
        index   : int | Sequence[int] | None
            Row position(s) to drop (by *position*, not label).
        inplace : bool, default False
            Mutate this DataSet or return a new one.

        Returns
        -------
        DataSet | None
        """
        # ------------------------------------------------------------------ #
        # 0. Normalise inputs
        # ------------------------------------------------------------------ #
        if isinstance(columns, str):
            columns = [columns]
        elif isinstance(index, int):
            index = [index]            
        if isinstance(index, (int, slice)):
            index = list(range(*index.indices(len(self.data))))  # handles slice
        elif index is not None and not isinstance(index, Sequence):
            raise TypeError("`index` must be int, slice or sequence of ints.")

        # ------------------------------------------------------------------ #
        # 1. Build working copy of rows
        # ------------------------------------------------------------------ #
        working_rows = deepcopy(self.data) if not inplace else self.data

        # ------------------------------------------------------------------ #
        # 2. Drop rows by position
        # ------------------------------------------------------------------ #
        if index:
            keep_mask    = [i not in index for i in range(len(working_rows))]
            working_rows = [r for r, keep in zip(working_rows, keep_mask) if keep]
            if hasattr(self, "_index") and self._index is not None:
                new_index = [self._index[i] for i in range(len(self._index)) if keep_mask[i]]
            else:
                new_index = None
        else:
            new_index = getattr(self, "_index", None)

        # ------------------------------------------------------------------ #
        # 3. Drop specified columns
        # ------------------------------------------------------------------ #
        if columns:
            # validation
            missing = [c for c in columns if c not in self.columns]
            if missing:
                raise KeyError(f"Column(s) not found: {missing}")
            working_rows = [
                {k: v for k, v in row.items() if k not in columns}
                for row in working_rows
            ]
            if getattr(self, "_index_name", None) in columns:
                self._index_name = None  # index column is gone

        # ------------------------------------------------------------------ #
        # 4. Finalise
        # ------------------------------------------------------------------ #
        if inplace:
            self.data    = working_rows
            self.columns = list(working_rows[0].keys()) if working_rows else []
            self._index  = new_index
            return None
        else:
            result             = Dataset(working_rows)
            result.columns     = list(working_rows[0].keys()) if working_rows else []
            result._index      = new_index
            result._index_name = getattr(self, "_index_name", None)
            return result  

    def rename(self, columns=None, inplace=False):
        """ Rename columns in the dataset.
        
        Parameters:
        ___________
            columns (dict): dictionary mapping old column names to new names
            inplace (bool): if true, modify the current dataset, otherwise return a new one
        
        Returns:
        ___________
            DataSet or None: new dataset with renamed columns, or None if inplace=True
            
        """ 
        if not columns:
            return self
        if not isinstance(columns, dict):
            raise TypeError("`columns` must be a dictionary mapping old column names to new names")
        
        new_data = []
        for row in self.data:
            new_row = {}
            for k, v in row.items():
                new_key = columns.get(k, k)
                new_row[new_key] = v
            new_data.append(new_row)

        if inplace:
            self.data = new_data
            self.columns = [columns.get(col, col) for col in self.columns]
            return None
        else:
            return Dataset(new_data)
        
    def groupby(self, by, sort=False):
        """
            Group the dataset by one more columns.

            Parameters:

                by (str or list of str): Column(s) to group by

            Returns:
                GroupBy: GroupBy object for aggregation
        """
        return GroupBy(self.data, by, sort)
    
    def sort_values(self, by, ascending=True, inplace=False, na_position='last'):
        if isinstance(by, str):
            by = [by]
        if isinstance(ascending, bool):
            ascending = [ascending] * len(by)

        if len(by) != len(ascending):
            raise ValueError("Length of 'by' and 'ascending' must match.")

        for col in by:
            if col not in self.columns:
                raise KeyError(f"Column '{col}' not found in dataset.")

        def invert(val):
            # For descending sort: convert to reverse-sortable proxy
            if val is None:
                return val
            if isinstance(val, (int, float)):
                return -val
            if isinstance(val, str):
                return ''.join(chr(255 - ord(c)) for c in val)
            return val  # fallback for other types

        def make_sort_key(row):
            key = []
            for col, asc in zip(by, ascending):
                val = row.get(col)
                is_nan = val is None

                # Handle NA marker
                nan_marker = 0 if is_nan and na_position == 'first' else 1 if is_nan else 2

                sort_val = val if asc else invert(val)
                key.append((nan_marker, sort_val))
            return tuple(key)

        sorted_data = sorted(self.data, key=make_sort_key)

        if inplace:
            self.data = sorted_data
            return None
        return Dataset(sorted_data)
    
    def len(self):
        return len(self.data)



from collections import defaultdict


class GroupBy:
    """
    GroupBy class for aggregating data based on specified keys.

    Supports:
    - agg (with named or dict-style aggregations)
    - sum, mean, min, max, count, first, last, size
    - apply (custom row-level logic)
    """

    def __init__(self, data, by, sort=False):
        self.by = by if isinstance(by, list) else [by]
        self.data = data
        self.sort = sort
        self.groups = self._group_data()

    def _group_data(self):
        grouped = defaultdict(list)
        for row in self.data:
            key = tuple(row[k] for k in self.by)
            grouped[key].append(row)
        if self.sort:
            return dict(sorted(grouped.items()))
        return grouped

    def _aggregate(self, agg_func_map, named_agg):
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            col_data = defaultdict(list)
            for row in rows:
                for col, val in row.items():
                    col_data[col].append(val)

            aggregated_row = {}

            if named_agg:
                # named agg: e.g. sales_mean=('sales', 'mean')
                for output_col, (input_col, agg_func) in agg_func_map.items():
                    values = col_data.get(input_col, [])
                    aggregated_row[output_col] = self._apply_func(agg_func, values, input_col)
            else:
                # dict agg: e.g. sales: ['sum', 'mean']
                for input_col, agg_funcs in agg_func_map.items():
                    if not isinstance(agg_funcs, list):
                        agg_funcs = [agg_funcs]
                    values = col_data.get(input_col, [])
                    for agg_func in agg_funcs:
                        result_col = self._output_col_name(input_col, agg_func)
                        aggregated_row[result_col] = self._apply_func(agg_func, values, input_col)

            for i, col in enumerate(self.by):
                aggregated_row[col] = group_key[i]

            result.append(aggregated_row)
        return Dataset(result)

    def _apply_func(self, func, values, input_col=None):
        if isinstance(func, str):
            if func == 'sum':
                return sum(values)
            elif func == 'mean' or func == 'avg':
                return sum(values) / len(values) if values else 0
            elif func == 'count':
                return len(values)
            elif func == 'min':
                return min(values) if values else None
            elif func == 'max':
                return max(values) if values else None
            elif func == 'first':
                return values[0] if values else None
            elif func == 'last':
                return values[-1] if values else None
            else:
                raise ValueError(f"Unknown aggregation function: {func}")
        elif callable(func):
            return func(values)
        else:
            raise TypeError(f"Aggregation function must be a string or callable, got {type(func)}")

    def _output_col_name(self, input_col, func):
        if isinstance(func, str):
            return f"{input_col}_{func}"
        elif callable(func):
            return f"{input_col}_{func.__name__}"
        else:
            raise TypeError("Aggregation function must be a string or callable")

    def agg(self, *args, **kwargs):
        if args and isinstance(args[0], dict):
            return self._aggregate(args[0], named_agg=False)
        elif kwargs:
            return self._aggregate(kwargs, named_agg=True)
        else:
            raise ValueError("agg() requires either a dict or named arguments")

    def sum(self):
        return self.agg({col: 'sum' for col in self._numeric_columns()})

    def mean(self):
        return self.agg({col: 'mean' for col in self._numeric_columns()})

    def avg(self):
        return self.agg({col: 'avg' for col in self._numeric_columns()})

    def min(self):
        return self.agg({col: 'min' for col in self._numeric_columns()})

    def max(self):
        return self.agg({col: 'max' for col in self._numeric_columns()})

    def count(self):
        return self.agg({col: 'count' for col in self.data[0].keys()})

    def first(self):
        return self.agg({col: 'first' for col in self.data[0].keys()})

    def last(self):
        return self.agg({col: 'last' for col in self.data[0].keys()})

    def size(self):
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            row = {col: key for col, key in zip(self.by, group_key)}
            row['size'] = len(rows)
            result.append(row)
        return Dataset(result)

    def apply(self, func):
        """
        Apply a custom function to each group.
        The function should return a dict or list of dicts.
        """
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            out = func(rows)
            if isinstance(out, dict):
                for i, col in enumerate(self.by):
                    out[col] = group_key[i]
                result.append(out)
            elif isinstance(out, list):
                for row in out:
                    for i, col in enumerate(self.by):
                        row[col] = group_key[i]
                    result.append(row)
            else:
                raise TypeError("apply() must return dict or list of dicts")
        return Dataset(result)

    def _numeric_columns(self):
        numeric_cols = set()
        for row in self.data:
            for col, val in row.items():
                if isinstance(val, (int, float)):
                    numeric_cols.add(col)
        return list(numeric_cols)
    
    def transform(self, func):
        """
        Apply a function to each group and broadcast result back to row level.
        The function should return a list of dicts with same length as group.
        """
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            out = func(rows)
            if not isinstance(out, list) or len(out) != len(rows):
                raise ValueError("transform() must return a list of dicts equal in length to input group")
            for i, col in enumerate(self.by):
                for row in out:
                    row[col] = group_key[i]
            result.extend(out)
        return Dataset(result)  

    def filter(self, func):
        """
        Keep only the groups for which the function returns True.
        """
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            if func(rows):
                for row in rows:
                    for i, col in enumerate(self.by):
                        row[col] = group_key[i]
                    result.append(row)
        return Dataset(result)      
    
    def describe(self):
        from .dataset import Dataset
        stats = ['count', 'mean', 'min', 'max']
        numeric_cols = self._numeric_columns()
        result = []
        for group_key, rows in self.groups.items():
            summary = {}
            for col in numeric_cols:
                values = [row[col] for row in rows if isinstance(row[col], (int, float))]
                summary[f'{col}_count'] = len(values)
                summary[f'{col}_mean'] = sum(values) / len(values) if values else None
                summary[f'{col}_min'] = min(values) if values else None
                summary[f'{col}_max'] = max(values) if values else None
            for i, col in enumerate(self.by):
                summary[col] = group_key[i]
            result.append(summary)
        return Dataset(result)    
    
    def cumsum(self):
        from .dataset import Dataset
        result = []
        numeric_cols = self._numeric_columns()
        for group_key, rows in self.groups.items():
            cum_sums = {col: 0 for col in numeric_cols}
            for row in rows:
                new_row = row.copy()
                for col in numeric_cols:
                    cum_sums[col] += row.get(col, 0)
                    new_row[f'{col}_cumsum'] = cum_sums[col]
                for i, col in enumerate(self.by):
                    new_row[col] = group_key[i]
                result.append(new_row)
        return Dataset(result)    
    
    def cumcount(self):
        from .dataset import Dataset
        result = []
        for group_key, rows in self.groups.items():
            for i, row in enumerate(rows):
                new_row = row.copy()
                new_row['cumcount'] = i
                for j, col in enumerate(self.by):
                    new_row[col] = group_key[j]
                result.append(new_row)
        return Dataset(result)    
    
    def rank(self):
        from .dataset import Dataset
        import numpy as np
        result = []
        numeric_cols = self._numeric_columns()
        for group_key, rows in self.groups.items():
            for col in numeric_cols:
                values = [row[col] for row in rows]
                sorted_indices = sorted(range(len(values)), key=lambda i: values[i])
                ranks = [0] * len(values)
                i = 0
                while i < len(values):
                    j = i
                    while j + 1 < len(values) and values[sorted_indices[j]] == values[sorted_indices[j + 1]]:
                        j += 1
                    rank_val = (i + j + 2) / 2  # average rank (1-based)
                    for k in range(i, j + 1):
                        ranks[sorted_indices[k]] = rank_val
                    i = j + 1
                for idx, row in enumerate(rows):
                    row[f'{col}_rank'] = ranks[idx]
            for row in rows:
                for i, col in enumerate(self.by):
                    row[col] = group_key[i]
                result.append(row)
        return Dataset(result)





class _LocIndexer:
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, key):
        from .dataset import Dataset
        from datetime import datetime
        from atrax import Series

        # handle (row_filter, col_filer) vs just row_filter
        if isinstance(key, tuple):
            row_filter, col_filter = key
        else:
            row_filter, col_filter = key, self.dataset.columns

        # normalize col_filter
        if isinstance(col_filter, str):
            col_filter = [col_filter]
        elif col_filter is None:
            col_filter = self.dataset.columns

        # row filtereing
        index = self.dataset._index
        data = self.dataset.data

        # single label
        if isinstance(row_filter, (str, int, float, datetime)):
            key_val = row_filter

            # smart datetime parsing if index is datetime-like
            if index and isinstance(index[0], datetime) and isinstance(row_filter, str):
                for fmt in ("%Y-%m-%d", "%m/%d/%Y", "%Y/%m/%d"):
                    try:
                        key_val = datetime.strptime(row_filter, fmt)
                        break
                    except ValueError:
                        continue
            filtered_data = [
                row for idx, row in zip(index, data)
                if idx == key_val
            ]
            new_index = [key_val] * len(filtered_data)


        # list of labels
        elif isinstance(row_filter, list) and all(not isinstance(x, bool) for x in row_filter):
            label_set = set(row_filter)
            filtered_data, new_index = zip(*[
                (row, idx) for idx, row in zip(index, data)
                if idx in label_set
            ]) if label_set else ([], [])

        # boolean mask
        elif isinstance(row_filter, list) and all(isinstance(x, bool) for x in row_filter):
            filtered_data = [row for row, keep in zip(data, row_filter) if keep]
            new_index = [idx for idx, keep in zip(index, row_filter) if keep]

        # callable
        elif callable(row_filter):
            filtered_data, new_index = zip(*[
                (row, idx) for idx, row in zip(index, data)
                if row_filter(row)
            ])

        # label slice (like 'a':'d')
        elif isinstance(row_filter, slice):
            # assume index is ordered and sliceable
            start, stop = row_filter.start, row_filter.stop
            filtered_data, new_index = zip(*[
                (row, idx) for idx, row in zip(index, data)
                if (start is None or idx >= start) and (stop is None or idx <= stop)
            ])

        # Boolean Series
        elif isinstance(row_filter, Series):
            bool_mask = row_filter.data
            filtered_data = [row for row, keep in zip(data, bool_mask) if keep]
            new_index = [idx for idx, keep in zip(index, bool_mask) if keep]

        else:
            # default: return all
            filtered_data = data
            new_index = index
        

        # column section
        result_data = [
            {col: row.get(col) for col in col_filter if col in row}
            for row in filtered_data
        ]

        return Dataset(result_data, index=list(new_index))
    
    def __setitem__(self, key, value):
        from atrax import Series


        if isinstance(key, tuple):
            row_filter, col_name = key
        else:
            row_filter = key
            col_name = None
        
        # get rows to modify
        if isinstance(row_filter, Series):
            bool_mask = row_filter.data
            rows = [row for row, keep in zip(self.dataset.data, bool_mask) if keep]

        elif isinstance(row_filter, list) and all(isinstance(b, bool) for b in row_filter):
            rows = [row for row, keep in zip(self.dataset.data, row_filter) if keep]

        elif callable(row_filter):
            rows = [row for row in self.dataset.data if row_filter(row)]

        else:
            raise TypeError("Invalid row selector for .loc")
        
        if col_name is None:
            raise ValueError("You must specify a column name for assignment with .loc")
        
        # apply assignment
        if isinstance(value, Series):
            if len(value.data) != len(rows):
                raise ValueError("Length of Series does not match number of selected rows")
            for row, val in zip(rows, value.data):
                row[col_name] = val
        elif isinstance(value, list):
            if len(value) != len(rows):
                raise ValueError("Length of list does not match number of selected rows")
            for row, val in zip(rows, value):
                row[col_name] = val
        elif callable(value):
            for row in rows:
                row[col_name] = value(row)
        else:
            for row in rows:
                row[col_name] = value  


        # Ensure the column exists in the dataset's columns list
        if col_name not in self.dataset.columns:
            self.dataset.columns.append(col_name)                     

    


class _iLocIndexer:
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, key):
        from .dataset import Dataset

        # normalize the key into (row_idx, col_idx)
        if isinstance(key, tuple): # ds.iloc[rows, cols]
            row_idx, col_idx = key
        else:                       # ds.iloc[rows] (cols=all)
            row_idx, col_idx = key, slice(None)

        # resolve rows
        if isinstance(row_idx, slice):
            rows = self.dataset.data[row_idx]
        elif isinstance(row_idx, (list, tuple)):
            rows = [self.dataset.data[i] for i in row_idx]
        else:
            rows = [self.dataset.data[row_idx]]

        # resolve columns
        if isinstance(col_idx, slice):
            column_names = self.dataset.columns[col_idx]
        elif isinstance(col_idx, (list, tuple)):
            column_names = [self.dataset.columns[i] for i in col_idx]
        else:
            column_names = [self.dataset.columns[col_idx]]

        # build filtered rows
        filtered = [
            {k: row[k] for k in column_names if k in row}
            for row in rows
        ]

        # get the index positions frmo the original dataset
        if isinstance(row_idx, slice):
            new_index = self.dataset._index[row_idx]
        elif isinstance(row_idx, (list, tuple)):
            new_index = [self.dataset._index[i] for i in row_idx]
        else:
            new_index = [self.dataset._index[row_idx]]

        return Dataset(filtered, index=new_index)




from datetime import datetime

class _DateTimeAccessor:
    def __init__(self, series):
        self.series = series

    def _convert(self, d, mode='date'):
        from .date_fns import parse_date
        if isinstance(d, str):
            date = parse_date(d)
            if mode == 'date':
                return date.weekday()
            elif mode == 'day':
                return date.day
            elif mode == 'month':
                return date.month
            elif mode == 'year':
                return date.year
            
        elif isinstance(d, datetime):
            if mode == 'date':
                return d.weekday()
            elif mode == 'day':
                return d.day
            elif mode == 'month':
                return d.month
            elif mode == 'year':
                return d.year
        else:
            raise TypeError(f"Unsupported type for date: {type(d)}")   

    @property
    def day(self):
        """
        Get the day of the month for each date in the Series.
        
        Returns:
        Series: A new Series with the day of the month.
        """
        from atrax.Series.series import Series

        return Series([self._convert(d, mode='day') for d in self.series.data], 
                      name=f"{self.series.name}_day", 
                      index=self.series.index)

    @property
    def month(self):
        """
        Get the month for each date in the Series.
        
        Returns:
        Series: A new Series with the month (1-12).
        """
        from atrax.Series.series import Series

        return Series([self._convert(d, mode='month') for d in self.series.data], 
                      name=f"{self.series.name}_month", 
                      index=self.series.index)     
    
    @property
    def year(self):
        """
        Get the year for each date in the Series.
        
        Returns:
        Series: A new Series with the year.
        """
        from atrax.Series.series import Series

        return Series([self._convert(d, mode='year') for d in self.series.data], 
                      name=f"{self.series.name}_year", 
                      index=self.series.index)

    @property
    def weekday(self):
        """
        Get the weekday of each date in the Series.
        
        Returns:
        Series: A new Series with the weekday (0=Monday, 6=Sunday).
        """
        from atrax.Series.series import Series

        return Series([self._convert(d) for d in self.series.data], 
                      name=f"{self.series.name}_weekday", 
                      index=self.series.index) 

    @property
    def is_weekend(self):
        """
        Check if each date in the Series is a weekend (Saturday or Sunday).
        
        Returns:
        Series: A new Series with int values indicating if the date is a weekend (1=yes, 0=no).
        """
        from atrax.Series.series import Series

        return Series([self._convert(d) >= 5 for d in self.series.data], 
                      name=f"{self.series.name}_is_weekend", 
                      index=self.series.index) 

    